"""Contains the shader source code for the rendering of the structures."""

from __future__ import annotations

from OpenGL.GL import (
    GL_FRAGMENT_SHADER,
    GL_GEOMETRY_SHADER,
    GL_VERTEX_SHADER,
    GLuint,
    glAttachShader,
    glCreateProgram,
    glLinkProgram,
)
from OpenGL.GL.shaders import ShaderCompilationError, compileShader

__copyright__ = "Copyright 2024, Molara"


def compile_shaders() -> list[GLuint]:
    """Compiles the shader program with the given shader source code in glsl.

    :return: The compiled shader program from pyopengl.
    """
    vertex_shader = compileShader(vertex_src_main, GL_VERTEX_SHADER)
    fragment_shader = compileShader(fragment_src_main, GL_FRAGMENT_SHADER)
    shader = glCreateProgram()
    glAttachShader(shader, vertex_shader)
    glAttachShader(shader, fragment_shader)
    glLinkProgram(shader)
    shaders: list[GLuint] = [shader]

    vertex_shader = compileShader(vertex_src_numbers, GL_VERTEX_SHADER)
    fragment_shader = compileShader(fragment_src_numbers, GL_FRAGMENT_SHADER)
    try:
        geometry_shader = compileShader(geometry_src_numbers_indices, GL_GEOMETRY_SHADER)
    except ShaderCompilationError:
        geometry_shader = compileShader(geometry_src_numbers_no_indices, GL_GEOMETRY_SHADER)
        print("Using geometry shader without indices")  # noqa: T201
    shader = glCreateProgram()
    glAttachShader(shader, vertex_shader)
    glAttachShader(shader, fragment_shader)
    glAttachShader(shader, geometry_shader)
    glLinkProgram(shader)
    shaders.append(shader)
    return shaders


vertex_src_main = """
# version 330 core

layout(location = 0) in vec3 a_position;
layout(location = 1) in vec3 a_normal;
layout(location = 2) in vec3 a_color;
layout(location = 3) in mat4 a_model;

uniform mat4 model;
uniform mat4 projection;
uniform mat4 view;
uniform vec3 light_direction;

out vec3 v_color;
out vec3 v_normal;
out vec3 v_light_dir;
out vec3 v_fragment_position;

void main()
{
    vec3 fragment_position = vec3(a_model * vec4(a_position, 1.0));
    v_fragment_position = fragment_position;
    gl_Position = projection * view * vec4(fragment_position, 1.0);
    v_color = a_color;
    v_light_dir = light_direction;
    v_normal = mat3(transpose(inverse(a_model))) * a_normal;
}
"""

fragment_src_main = """
# version 330 core

in vec3 v_color;
in vec3 v_light_dir;
in vec3 v_normal;
in vec3 v_fragment_position;

uniform vec3 camera_position;

out vec4 out_color;

void main()
{
    vec3 normal = normalize(v_normal);
    vec3 light_dir = normalize(-v_light_dir);
    vec3 light_fragment_direction = normalize(light_dir - v_fragment_position);
    vec3 camera_fragment_direction = normalize(camera_position - v_fragment_position);
    vec3 halfway = normalize(light_fragment_direction + camera_fragment_direction);


    float diff = max(dot(normal, light_dir), 0.0) * 0.66667 + 0.33334;

    float spec = pow(max(dot(normal, halfway), 0.0), 25);

    vec3 light_color = vec3(1.0, 1.0, 1.0);    // Light color

    vec3 result = v_color * light_color * diff;// + spec;
    out_color = vec4(result, 1.0);
}
"""

vertex_src_numbers = """
#version 410 core
layout (location = 0) in uint digit;
layout (location = 1) in vec3 position;

uniform float aspect_ratio;
uniform float scale;
uniform mat4 projection;
uniform mat4 view;

out float aspect_ratio_v;
out uint digit_v;
out float scale_v;

void main()
{
    aspect_ratio_v = aspect_ratio;
    digit_v = digit;
    scale_v = scale;
    gl_Position = projection * view * vec4(position, 1.0);
}
"""

fragment_src_numbers = """
#version 410 core
uniform vec3 color_in;

out vec4 out_color;

void main()
{
    out_color = vec4(color_in, 1.0);
}
"""

geometry_src_numbers_indices = """
#version 330 core
layout (points) in;
layout (triangle_strip, max_vertices = 350) out;

in float aspect_ratio_v[];
in uint digit_v[];
in float scale_v[];

void draw_horizontal_line(vec4 position, float aspect_ratio, float scale)
{
    float radius = 0.05 * scale;
    float angle = 0.0;
    int subsections = 2;
    float pi = 3.14159265359;
    float pi_2 = pi / 2;
    float inv_ar = 1 / aspect_ratio;
    float x = 0.2 * inv_ar * scale;
    float y = 0.05 * scale;

    gl_Position = position + vec4(-x, -y, 0.0, 0.0);    // 1:bottom-left
    EmitVertex();
    gl_Position = position + vec4( x, -y, 0.0, 0.0);    // 2:bottom-right
    EmitVertex();
    gl_Position = position + vec4(-x,  y, 0.0, 0.0);    // 3:top-left
    EmitVertex();
    gl_Position = position + vec4( x,  y, 0.0, 0.0);    // 4:top-right
    EmitVertex();
    vec4 temp_pos = position + vec4(x, 0.0, 0.0, 0.0);
    gl_Position = temp_pos;
    EmitVertex();

    for (int i = 0; i <= subsections; i++)
    {
        gl_Position = temp_pos + vec4(radius * cos(angle - pi_2) * inv_ar, radius * sin(angle - pi_2), 0.0, 0.0);
        EmitVertex();
        gl_Position = temp_pos;
        EmitVertex();
        angle += 3.14159265359 / subsections;
    }
    EndPrimitive();

    temp_pos = position + vec4(-x, 0.0, 0.0, 0.0);
    gl_Position = temp_pos;
    EmitVertex();

    angle = 0.0;
    for (int i = 0; i <= subsections; i++)
    {
        gl_Position = temp_pos + vec4(radius * cos(angle + pi_2) * inv_ar, radius * sin(angle + pi_2), 0.0, 0.0);
        EmitVertex();
        gl_Position = temp_pos;
        EmitVertex();
        angle += 3.14159265359 / subsections;
    }
    EndPrimitive();
}
void draw_vertical_line(vec4 position_, float aspect_ratio, float scale)
{
    float radius = 0.05 * scale;
    float angle = 0.0;
    int subsections = 2;
    float pi = 3.14159265359;
    float pi_2 = pi / 2;
    float inv_ar = 1 / aspect_ratio;
    float x = 0.05 * inv_ar * scale;
    float y = 0.2 * scale;

    gl_Position = position_ + vec4(-x, -y, 0.0, 0.0);    // 1:bottom-left
    EmitVertex();
    gl_Position = position_ + vec4( x, -y, 0.0, 0.0);    // 2:bottom-right
    EmitVertex();
    gl_Position = position_ + vec4(-x,  y, 0.0, 0.0);    // 3:top-left
    EmitVertex();
    gl_Position = position_ + vec4( x,  y, 0.0, 0.0);    // 4:top-right
    EmitVertex();
    vec4 temp_pos = position_ + vec4(0.0, y, 0.0, 0.0);
    gl_Position = temp_pos;
    EmitVertex();

    for (int i = 0; i <= subsections; i++)
    {
        gl_Position = temp_pos + vec4(radius * cos(angle) * inv_ar, radius * sin(angle), 0.0, 0.0);
        EmitVertex();
        gl_Position = temp_pos;
        EmitVertex();
        angle += 3.14159265359 / subsections;
    }
    EndPrimitive();

    temp_pos = position_ + vec4(0.0, -y, 0.0, 0.0);
    gl_Position = temp_pos;
    EmitVertex();

    angle = 0.0;
    for (int i = 0; i <= subsections; i++)
    {
        gl_Position = temp_pos + vec4(radius * cos(angle) * inv_ar, radius * sin(angle), 0.0, 0.0);
        EmitVertex();
        gl_Position = temp_pos;
        EmitVertex();
        angle -= 3.14159265359 / subsections;
    }
    EndPrimitive();
}

void display_0(vec4 position, float aspect_ratio, float scale)
{
    float inv_ar = 1 / aspect_ratio;
    vec4 temp_pos = vec4(position.x, position.y + 0.4 * scale, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4(position.x, position.y - 0.4 * scale, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x - 0.2 * scale * inv_ar), position.y + 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x + 0.2 * scale * inv_ar), position.y + 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x - 0.2 * scale * inv_ar), position.y - 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x + 0.2 * scale * inv_ar), position.y - 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);
}

void display_1(vec4 position, float aspect_ratio, float scale)
{
    float inv_ar = 1 / aspect_ratio;

    vec4 temp_pos = vec4((position.x + 0.2 * scale * inv_ar), position.y + 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x + 0.2 * scale * inv_ar), position.y - 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);
}

void display_2(vec4 position, float aspect_ratio, float scale)
{
    float inv_ar = 1 / aspect_ratio;
    vec4 temp_pos = vec4(position.x * 1.0, position.y + 0.4 * scale, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4(position.x* 1.0, position.y - 0.4 * scale, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4(position.x* 1.0, position.y, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x + 0.2 * scale * inv_ar), position.y + 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x - 0.2 * scale * inv_ar), position.y - 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);
}

void display_3(vec4 position, float aspect_ratio, float scale)
{
    float inv_ar = 1 / aspect_ratio;
    vec4 temp_pos = vec4(position.x, position.y + 0.4 * scale, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4(position.x, position.y - 0.4 * scale, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4(position.x, position.y, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x + 0.2 * scale * inv_ar), position.y + 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x + 0.2 * scale * inv_ar), position.y - 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);
}

void display_4(vec4 position, float aspect_ratio, float scale)
{
    float inv_ar = 1 / aspect_ratio;
    vec4 temp_pos = vec4(position.x, position.y, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);


    temp_pos = vec4((position.x - 0.2 * scale * inv_ar), position.y + 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x + 0.2 * scale * inv_ar), position.y + 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x + 0.2 * scale * inv_ar), position.y - 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);
}

void display_5(vec4 position, float aspect_ratio, float scale)
{
    float inv_ar = 1 / aspect_ratio;
    vec4 temp_pos = vec4(position.x, position.y + 0.4 * scale, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4(position.x, position.y - 0.4 * scale, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4(position.x, position.y, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x - 0.2 * scale * inv_ar), position.y + 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x + 0.2 * scale * inv_ar), position.y - 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);
}

void display_6(vec4 position, float aspect_ratio, float scale)
{
    float inv_ar = 1 / aspect_ratio;
    vec4 temp_pos = vec4(position.x, position.y + 0.4 * scale, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4(position.x, position.y - 0.4 * scale, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4(position.x, position.y, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x - 0.2 * scale * inv_ar), position.y + 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x - 0.2 * scale * inv_ar), position.y - 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x + 0.2 * scale * inv_ar), position.y - 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);
}

void display_7(vec4 position, float aspect_ratio, float scale)
{
    float inv_ar = 1 / aspect_ratio;

    vec4 temp_pos = vec4(position.x, position.y + 0.4 * scale, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x + 0.2 * scale * inv_ar), position.y + 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x + 0.2 * scale * inv_ar), position.y - 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);
}

void display_8(vec4 position, float aspect_ratio, float scale)
{
    float inv_ar = 1 / aspect_ratio;
    vec4 temp_pos = vec4(position.x, position.y + 0.4 * scale, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4(position.x, position.y - 0.4 * scale, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4(position.x, position.y, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x - 0.2 * scale * inv_ar), position.y + 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x + 0.2 * scale * inv_ar), position.y + 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x - 0.2 * scale * inv_ar), position.y - 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x + 0.2 * scale * inv_ar), position.y - 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);
}

void display_9(vec4 position, float aspect_ratio, float scale)
{
    float inv_ar = 1 / aspect_ratio;
    vec4 temp_pos = vec4(position.x, position.y + 0.4 * scale, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4(position.x, position.y - 0.4 * scale, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4(position.x, position.y, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x - 0.2 * scale * inv_ar), position.y + 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x + 0.2 * scale * inv_ar), position.y + 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x + 0.2 * scale * inv_ar), position.y - 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);
}

void display_digit(vec4 position, float aspect_ratio, uint digit, float scale)
{
    if (digit == 0u){
        display_0(position, aspect_ratio, scale);
    }
    else if (digit == 1u){
        display_1(position, aspect_ratio, scale);
    }
    else if (digit == 2u){
        display_2(position, aspect_ratio, scale);
    }
    else if (digit == 3u){
        display_3(position, aspect_ratio, scale);
    }
    else if (digit == 4u){
        display_4(position, aspect_ratio, scale);
    }
    else if (digit == 5u){
        display_5(position, aspect_ratio, scale);
    }
    else if (digit == 6u){
        display_6(position, aspect_ratio, scale);
    }
    else if (digit == 7u){
        display_7(position, aspect_ratio, scale);
    }
    else if (digit == 8u){
        display_8(position, aspect_ratio, scale);
    }
    else if (digit == 9u){
        display_9(position, aspect_ratio, scale);
    }
}

void main() {
    float ooo = 0.0;
    uint digit = digit_v[0];
    if (digit < 10u){
        display_digit(gl_in[0].gl_Position, aspect_ratio_v[0], digit, scale_v[0]);
    }
    else if (digit < 100u){
        uint digit1 = digit / 10u;
        display_digit(gl_in[0].gl_Position - vec4(0.2,0.0,0.0,0.0) * scale_v[0], aspect_ratio_v[0], digit1, scale_v[0]);
        uint digit2 = digit % 10u;
        display_digit(gl_in[0].gl_Position + vec4(0.2,0.0,0.0,0.0) * scale_v[0], aspect_ratio_v[0], digit2, scale_v[0]);
    }
    else if (digit < 1000u){
        uint digit1 = digit / 100u;
        display_digit(gl_in[0].gl_Position - vec4(0.4,0.0,0.0,0.0) * scale_v[0], aspect_ratio_v[0], digit1, scale_v[0]);
        uint digit2 = (digit % 100u) / 10u;
        display_digit(gl_in[0].gl_Position, aspect_ratio_v[0], digit2, scale_v[0]);
        uint digit3 = digit % 10u;
        display_digit(gl_in[0].gl_Position + vec4(0.4,0.0,0.0,0.0) * scale_v[0], aspect_ratio_v[0], digit3, scale_v[0]);
    }
}
"""

geometry_src_numbers_no_indices = """
#version 330 core
layout (points) in;
layout (triangle_strip, max_vertices = 100) out;

in float aspect_ratio_v[];
in uint digit_v[];
in float scale_v[];

void draw_horizontal_line(vec4 position, float aspect_ratio, float scale)
{
    float radius = 0.05 * scale;
    float angle = 0.0;
    int subsections = 2;
    float pi = 3.14159265359;
    float pi_2 = pi / 2;
    float inv_ar = 1 / aspect_ratio;
    float x = 0.2 * inv_ar * scale;
    float y = 0.05 * scale;

    gl_Position = position + vec4(-x, -y, 0.0, 0.0);    // 1:bottom-left
    EmitVertex();
    gl_Position = position + vec4( x, -y, 0.0, 0.0);    // 2:bottom-right
    EmitVertex();
    gl_Position = position + vec4(-x,  y, 0.0, 0.0);    // 3:top-left
    EmitVertex();
    gl_Position = position + vec4( x,  y, 0.0, 0.0);    // 4:top-right
    EmitVertex();
    vec4 temp_pos = position + vec4(x, 0.0, 0.0, 0.0);
    gl_Position = temp_pos;
    EmitVertex();

    for (int i = 0; i <= subsections; i++)
    {
        gl_Position = temp_pos + vec4(radius * cos(angle - pi_2) * inv_ar, radius * sin(angle - pi_2), 0.0, 0.0);
        EmitVertex();
        gl_Position = temp_pos;
        EmitVertex();
        angle += 3.14159265359 / subsections;
    }
    EndPrimitive();

    temp_pos = position + vec4(-x, 0.0, 0.0, 0.0);
    gl_Position = temp_pos;
    EmitVertex();

    angle = 0.0;
    for (int i = 0; i <= subsections; i++)
    {
        gl_Position = temp_pos + vec4(radius * cos(angle + pi_2) * inv_ar, radius * sin(angle + pi_2), 0.0, 0.0);
        EmitVertex();
        gl_Position = temp_pos;
        EmitVertex();
        angle += 3.14159265359 / subsections;
    }
    EndPrimitive();
}
void draw_vertical_line(vec4 position_, float aspect_ratio, float scale)
{
    float radius = 0.05 * scale;
    float angle = 0.0;
    int subsections = 2;
    float pi = 3.14159265359;
    float pi_2 = pi / 2;
    float inv_ar = 1 / aspect_ratio;
    float x = 0.05 * inv_ar * scale;
    float y = 0.2 * scale;

    gl_Position = position_ + vec4(-x, -y, 0.0, 0.0);    // 1:bottom-left
    EmitVertex();
    gl_Position = position_ + vec4( x, -y, 0.0, 0.0);    // 2:bottom-right
    EmitVertex();
    gl_Position = position_ + vec4(-x,  y, 0.0, 0.0);    // 3:top-left
    EmitVertex();
    gl_Position = position_ + vec4( x,  y, 0.0, 0.0);    // 4:top-right
    EmitVertex();
    vec4 temp_pos = position_ + vec4(0.0, y, 0.0, 0.0);
    gl_Position = temp_pos;
    EmitVertex();

    for (int i = 0; i <= subsections; i++)
    {
        gl_Position = temp_pos + vec4(radius * cos(angle) * inv_ar, radius * sin(angle), 0.0, 0.0);
        EmitVertex();
        gl_Position = temp_pos;
        EmitVertex();
        angle += 3.14159265359 / subsections;
    }
    EndPrimitive();

    temp_pos = position_ + vec4(0.0, -y, 0.0, 0.0);
    gl_Position = temp_pos;
    EmitVertex();

    angle = 0.0;
    for (int i = 0; i <= subsections; i++)
    {
        gl_Position = temp_pos + vec4(radius * cos(angle) * inv_ar, radius * sin(angle), 0.0, 0.0);
        EmitVertex();
        gl_Position = temp_pos;
        EmitVertex();
        angle -= 3.14159265359 / subsections;
    }
    EndPrimitive();
}

void display_0(vec4 position, float aspect_ratio, float scale)
{
    float inv_ar = 1 / aspect_ratio;
    vec4 temp_pos = vec4(position.x, position.y + 0.4 * scale, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4(position.x, position.y - 0.4 * scale, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x - 0.2 * scale * inv_ar), position.y + 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x + 0.2 * scale * inv_ar), position.y + 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x - 0.2 * scale * inv_ar), position.y - 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x + 0.2 * scale * inv_ar), position.y - 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);
}

void display_1(vec4 position, float aspect_ratio, float scale)
{
    float inv_ar = 1 / aspect_ratio;

    vec4 temp_pos = vec4((position.x + 0.2 * scale * inv_ar), position.y + 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x + 0.2 * scale * inv_ar), position.y - 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);
}

void display_2(vec4 position, float aspect_ratio, float scale)
{
    float inv_ar = 1 / aspect_ratio;
    vec4 temp_pos = vec4(position.x * 1.0, position.y + 0.4 * scale, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4(position.x* 1.0, position.y - 0.4 * scale, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4(position.x* 1.0, position.y, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x + 0.2 * scale * inv_ar), position.y + 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x - 0.2 * scale * inv_ar), position.y - 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);
}

void display_3(vec4 position, float aspect_ratio, float scale)
{
    float inv_ar = 1 / aspect_ratio;
    vec4 temp_pos = vec4(position.x, position.y + 0.4 * scale, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4(position.x, position.y - 0.4 * scale, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4(position.x, position.y, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x + 0.2 * scale * inv_ar), position.y + 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x + 0.2 * scale * inv_ar), position.y - 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);
}

void display_4(vec4 position, float aspect_ratio, float scale)
{
    float inv_ar = 1 / aspect_ratio;
    vec4 temp_pos = vec4(position.x, position.y, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);


    temp_pos = vec4((position.x - 0.2 * scale * inv_ar), position.y + 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x + 0.2 * scale * inv_ar), position.y + 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x + 0.2 * scale * inv_ar), position.y - 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);
}

void display_5(vec4 position, float aspect_ratio, float scale)
{
    float inv_ar = 1 / aspect_ratio;
    vec4 temp_pos = vec4(position.x, position.y + 0.4 * scale, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4(position.x, position.y - 0.4 * scale, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4(position.x, position.y, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x - 0.2 * scale * inv_ar), position.y + 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x + 0.2 * scale * inv_ar), position.y - 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);
}

void display_6(vec4 position, float aspect_ratio, float scale)
{
    float inv_ar = 1 / aspect_ratio;
    vec4 temp_pos = vec4(position.x, position.y + 0.4 * scale, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4(position.x, position.y - 0.4 * scale, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4(position.x, position.y, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x - 0.2 * scale * inv_ar), position.y + 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x - 0.2 * scale * inv_ar), position.y - 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x + 0.2 * scale * inv_ar), position.y - 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);
}

void display_7(vec4 position, float aspect_ratio, float scale)
{
    float inv_ar = 1 / aspect_ratio;

    vec4 temp_pos = vec4(position.x, position.y + 0.4 * scale, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x + 0.2 * scale * inv_ar), position.y + 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x + 0.2 * scale * inv_ar), position.y - 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);
}

void display_8(vec4 position, float aspect_ratio, float scale)
{
    float inv_ar = 1 / aspect_ratio;
    vec4 temp_pos = vec4(position.x, position.y + 0.4 * scale, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4(position.x, position.y - 0.4 * scale, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4(position.x, position.y, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x - 0.2 * scale * inv_ar), position.y + 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x + 0.2 * scale * inv_ar), position.y + 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x - 0.2 * scale * inv_ar), position.y - 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x + 0.2 * scale * inv_ar), position.y - 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);
}

void display_9(vec4 position, float aspect_ratio, float scale)
{
    float inv_ar = 1 / aspect_ratio;
    vec4 temp_pos = vec4(position.x, position.y + 0.4 * scale, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4(position.x, position.y - 0.4 * scale, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4(position.x, position.y, position.z, position.w);
    draw_horizontal_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x - 0.2 * scale * inv_ar), position.y + 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x + 0.2 * scale * inv_ar), position.y + 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);

    temp_pos = vec4((position.x + 0.2 * scale * inv_ar), position.y - 0.2*scale, position.z, position.w);
    draw_vertical_line(temp_pos, aspect_ratio, scale);
}

void display_digit(vec4 position, float aspect_ratio, uint digit, float scale)
{
    if (digit == 0u){
        display_0(position, aspect_ratio, scale);
    }
    else if (digit == 1u){
        display_1(position, aspect_ratio, scale);
    }
    else if (digit == 2u){
        display_2(position, aspect_ratio, scale);
    }
    else if (digit == 3u){
        display_3(position, aspect_ratio, scale);
    }
    else if (digit == 4u){
        display_4(position, aspect_ratio, scale);
    }
    else if (digit == 5u){
        display_5(position, aspect_ratio, scale);
    }
    else if (digit == 6u){
        display_6(position, aspect_ratio, scale);
    }
    else if (digit == 7u){
        display_7(position, aspect_ratio, scale);
    }
    else if (digit == 8u){
        display_8(position, aspect_ratio, scale);
    }
    else if (digit == 9u){
        display_9(position, aspect_ratio, scale);
    }
}

void main() {
    float ooo = 0.0;
    uint digit = digit_v[0];
}
"""

geometry_src_lines_with_linewidth = """
#version 330 core
layout(lines) in;
layout(triangle_strip, max_vertices = 4) out;

uniform float lineWidth;

void main() {
    vec4 p0 = gl_in[0].gl_Position;
    vec4 p1 = gl_in[1].gl_Position;

    vec2 dir = normalize(p1.xy - p0.xy);
    vec2 normal = vec2(-dir.y, dir.x) * lineWidth * 0.5;

    gl_Position = p0 + vec4(normal, 0.0, 0.0);
    EmitVertex();

    gl_Position = p1 + vec4(normal, 0.0, 0.0);
    EmitVertex();

    gl_Position = p0 - vec4(normal, 0.0, 0.0);
    EmitVertex();

    gl_Position = p1 - vec4(normal, 0.0, 0.0);
    EmitVertex();

    EndPrimitive();
}
"""



"""Function for rendering atom labels."""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from molara.rendering.camera import Camera
    from molara.structure.structure import Structure


def init_atom_number(structure: Structure) -> tuple[np.ndarray, np.ndarray]:
    """Initialize the labels for numbering atoms.

    :param structure: Structure object
    """
    number_of_atoms = len(structure.atoms)

    # Initialize the arrays for the atom numbers.
    positions_3d = np.zeros((number_of_atoms, 3), dtype=np.float32)
    digits = np.zeros(number_of_atoms, dtype=np.int32)

    return digits, positions_3d


def calculate_atom_number_arrays(
    digits: np.ndarray,
    positions_3d: np.ndarray,
    structure: Structure,
    camera: Camera,
) -> None:
    """Calculate the arrays for the atom numbers.

    :param digits: The digits to be displayed.
    :param positions_3d: The 3D positions of the atoms.
    :param structure: Structure object
    :param camera: Camera object
    """
    number_of_atoms = len(structure.atoms)

    i = 0
    while i < number_of_atoms:
        atom_position_3d = structure.atoms[i].position

        atom_cam_vec = camera.position - atom_position_3d
        atom_cam_vec /= np.linalg.norm(atom_cam_vec)
        position_number_3d = (
            atom_position_3d
            + atom_cam_vec
            * structure.drawer.sphere_default_radius
            * structure.drawer.sphere_scale
            * structure.atoms[i].vdw_radius
            * 1.75
        )
        positions_3d[i] = position_number_3d
        digits[i] = i + 1

        i += 1



"""Camera class for the opengl widget."""

from __future__ import annotations

import json
from typing import Literal

import numpy as np
import numpy.typing as npt
import pyrr

__copyright__ = "Copyright 2024, Molara"


class Camera:
    """Creates a Camera object."""

    def __init__(self, width: float, height: float) -> None:
        """Create a Camera object.

        :param width: Width of the opengl widget.
        :param height: Height of the opengl widget.
        """
        self.fov = 45.0
        self.width = width
        self.height = height
        self.near = 0.1
        self.far = 100.0
        self.position = pyrr.Vector3([1.0, 0.0, 0.0], dtype=np.float32)
        self.up_vector = pyrr.Vector3([0.0, 1.0, 0.0], dtype=np.float32)
        self.right_vector = pyrr.Vector3([0.0, 0.0, -1.0], dtype=np.float32)
        self.distance_from_target = 5.0
        self.zoom_sensitivity = 0.15
        self.projection_matrix: npt.ArrayLike | None = None
        self.orthographic_projection = False  # specify whether orthographic projection shall be used
        self.calculate_projection_matrix()

        self.rotation = pyrr.Quaternion()
        self.translation = pyrr.Vector3([0.0, 0.0, 0.0], dtype=np.float32)
        self.position *= self.distance_from_target
        self.target = pyrr.Vector3([0.0, 0.0, 0.0], dtype=np.float32)

        self.last_rotation = self.rotation
        self.last_translation = self.translation
        self.initial_target = self.target
        self.initial_position = pyrr.Vector3(
            pyrr.vector3.normalize(self.position),
            dtype=np.float32,
        )
        self.initial_up_vector = self.up_vector
        self.initial_right_vector = self.right_vector
        self.view_matrix = pyrr.matrix44.create_look_at(
            self.position,
            self.target,
            self.up_vector,
            dtype=np.float32,
        )
        self.view_matrix_inv = pyrr.matrix44.inverse(self.view_matrix)

    def calculate_projection_matrix(self) -> None:
        """Calculate the projection matrix to get from world to camera space."""
        width, height = self.width, self.height  # Width and height of the opengl widget
        if self.orthographic_projection:
            # calculate width and height of the clipping plane
            # such that it matches the field of view in the perspective projection
            h = self.distance_from_target * np.tan(np.radians(self.fov / 2))
            w = h * width / height
            self.projection_matrix = pyrr.matrix44.create_orthogonal_projection_matrix(
                -w,
                w,
                -h,
                h,
                self.near,
                self.far,
                dtype=np.float32,
            )
            return
        self.projection_matrix = pyrr.matrix44.create_perspective_projection_matrix(
            self.fov,
            width / height,
            self.near,
            self.far,
            dtype=np.float32,
        )
        self.projection_matrix_inv = pyrr.matrix44.inverse(self.projection_matrix)

    def reset(
        self,
        width: float,
        height: float,
        dy_struct: float | None = None,
        dz_struct: float | None = None,
    ) -> None:
        """Reset the camera.

        :param width: Width of the opengl widget.
        :param height: Height of the opengl widget.
        :param dy_struct: extent of the structure in y-direction
        :param dz_struct: extent of the structure in z-direction
        """
        self.width = width
        self.height = height

        # set distance from target such that the structure fits into the view
        # equation to be solved:
        # tan(fov [°] / 2)) * distance_from_target = dy / 2
        # if vertical (y) length is limiting, dy = dy_struct. Otherwise, dy = height/width*dz_struct
        # z-axis is horizontal by default.
        distance_from_target = None
        if dy_struct and dz_struct:
            extra_space_factor = 1.5
            dy = dy_struct if dy_struct * self.width > dz_struct * self.height else dz_struct * self.height / self.width
            distance_from_target = extra_space_factor * dy / (2 * np.tan(np.radians(self.fov / 2)))

        self.set_position(
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.0, 0.0, -1.0],
            distance_from_target,
        )

    def set_position(
        self,
        position: list[float],
        up_vector: list[float],
        right_vector: list[float],
        distance_from_target: float | None = None,
    ) -> None:
        """Set camera position and orientation.

        :param position: coordinates of camera position
        :param up_vector: vector indicating which direction is upward for the camera
        :param right_vector: vector indicating which direction is right for the camera
        :param distance_from_target: distance between the camera and its target
        """
        self.position = pyrr.Vector3(position, dtype=np.float32)
        self.up_vector = pyrr.Vector3(up_vector, dtype=np.float32)
        self.right_vector = pyrr.Vector3(right_vector, dtype=np.float32)
        self.distance_from_target = distance_from_target if distance_from_target else self.distance_from_target
        self.projection_matrix = None
        self.calculate_projection_matrix()

        self.rotation = pyrr.Quaternion()
        self.last_rotation = self.rotation
        self.translation = pyrr.Vector3([0.0, 0.0, 0.0], dtype=np.float32)
        self.last_translation = self.translation
        self.position *= self.distance_from_target
        self.target = pyrr.Vector3([0.0, 0.0, 0.0], dtype=np.float32)
        self.initial_target = self.target
        self.initial_position = pyrr.Vector3(
            pyrr.vector3.normalize(self.position),
            dtype=np.float32,
        )
        self.initial_up_vector = self.up_vector
        self.initial_right_vector = self.right_vector

        self.view_matrix = pyrr.matrix44.create_look_at(
            self.position,
            self.target,
            self.up_vector,
            dtype=np.float32,
        )
        self.view_matrix_inv = pyrr.matrix44.inverse(self.view_matrix)
        self.projection_matrix_inv = pyrr.matrix44.inverse(self.projection_matrix)

    def set_rotation(self, axis: Literal["x", "y", "z"]) -> None:
        """Align the camera rotation with one of the major axes ("x", "y", "z").

        :param axis: specifies along which axis camera view shall be aligned.
        """
        self.initial_position = pyrr.Vector3([1.0, 0.0, 0.0], dtype=np.float32)
        self.initial_up_vector = pyrr.Vector3([0.0, 1.0, 0.0], dtype=np.float32)
        self.initial_right_vector = pyrr.Vector3([0.0, 0.0, -1.0], dtype=np.float32)
        if axis == "x":
            rotation_axis = pyrr.Vector3([0, 1, 0], dtype=np.float32)
            rotation_angle = 0.0
        elif axis == "y":
            rotation_axis = pyrr.Vector3([0, 0, 1], dtype=np.float32)
            rotation_angle = np.pi / 2.0
        elif axis == "z":
            rotation_axis = pyrr.Vector3([0, 1, 0], dtype=np.float32)
            rotation_angle = np.pi / 2.0
        self.rotation = pyrr.Quaternion() * pyrr.Quaternion.from_axis_rotation(
            rotation_axis,
            rotation_angle,
        )
        self.last_rotation = self.rotation
        self.update()

    def center_coordinates(self) -> None:
        """Reset camera translation such that center of structure is in center of view."""
        self.translation = pyrr.Vector3([0.0, 0.0, 0.0], dtype=np.float32)
        self.last_translation = self.translation
        self.update()

    def set_distance_from_target(self, num_steps: int) -> None:
        """Set the distance between the camera and its target.

        :param num_steps: Number of zoom steps. From this, factor is calculated that is multiplied
            with the normalized camera position vector and the current distance
            between the camera and the target.
        """
        zoom_factor = 1.0
        zoom_factor += num_steps * 1  # Empirical value to control zoom sensitivity
        zoom_factor = max(
            0.1,
            zoom_factor,
        )  # Limit zoom factor to avoid zooming too far
        self.distance_from_target += np.log10(zoom_factor * self.zoom_sensitivity) * np.sign(zoom_factor - 1)
        self.distance_from_target = max(self.distance_from_target, 1.0)

    def toggle_projection(self) -> None:
        """Toggle between perspective and orthographic projection."""
        self.orthographic_projection = not self.orthographic_projection
        self.calculate_projection_matrix()

    def update(self, save: bool = False) -> None:
        """Update the camera position and orientation.

        :param save: specifies whether updated camera orientation shall be saved as new reference
        """
        self.up_vector = self.rotation * self.initial_up_vector
        self.right_vector = self.rotation * self.initial_right_vector
        self.position = self.rotation * self.initial_position * self.distance_from_target + self.translation
        self.position = np.array(self.position, dtype=np.float32)
        self.target = self.initial_target + self.translation
        if save:
            self.last_rotation = self.rotation
            self.last_translation = self.translation
        self.view_matrix = pyrr.matrix44.create_look_at(
            self.position,
            self.target,
            self.up_vector,
            dtype=np.float32,
        )
        self.view_matrix_inv = pyrr.matrix44.inverse(self.view_matrix)
        if self.orthographic_projection:
            self.calculate_projection_matrix()

    def set_translation_vector(
        self,
        old_mouse_position: np.ndarray,
        mouse_position: np.ndarray,
    ) -> None:
        """Calculate the translation matrix using the normalized mouse positions.

        :param old_mouse_position: Old normalized x and y coordinate of the mouse position on the opengl widget.
        :param mouse_position: New normalized x and y coordinate of the mouse position on the opengl widget.
        """
        x_translation = -(mouse_position[0] - old_mouse_position[0])
        y_translation = -(mouse_position[1] - old_mouse_position[1])
        self.translation = self.right_vector * x_translation + self.up_vector * y_translation + self.last_translation

    def set_rotation_quaternion(
        self,
        old_mouse_position: np.ndarray,
        new_mouse_position: np.ndarray,
    ) -> None:
        """Calculate the rotation quaternion using the normalized mouse positions.

        :param old_mouse_position: Old normalized x and y coordinate of the mouse position on the opengl widget.
        :param new_mouse_position: New normalized x and y coordinate of the mouse position on the opengl widget.
        """

        def calculate_arcball_point(x: float, y: float) -> np.ndarray:
            """Calculate the x, y, and z on the surface of an invisible sphere.

            :param x: Normalized x coordinate.
            :param y: Normalized y coordinate.
            :return: x, y, and z on the surface of an invisible sphere.
            """
            z = x
            squared_sum = z**2 + y**2
            if squared_sum <= 1:
                x = np.sqrt(1 - squared_sum)
            else:
                x = 0.0
                length = np.sqrt(squared_sum)
                z /= length
                y /= length
            return np.array([x, y, -z], dtype=np.float32)

        previous_arcball_point = calculate_arcball_point(
            old_mouse_position[0],
            old_mouse_position[1],
        )
        current_arcball_point = calculate_arcball_point(
            new_mouse_position[0],
            new_mouse_position[1],
        )

        tolerance_parallel = 1e-5
        if np.linalg.norm(previous_arcball_point - current_arcball_point) > tolerance_parallel:
            rotation_axis = np.cross(current_arcball_point, previous_arcball_point)
            rotation_axis = pyrr.vector3.normalize(rotation_axis)
            rotation_angle = np.arccos(
                np.clip(np.dot(previous_arcball_point, current_arcball_point), -1, 1),
            )

            self.rotation = self.last_rotation * pyrr.Quaternion.from_axis_rotation(
                rotation_axis,
                rotation_angle,
            )
        else:
            self.rotation = self.last_rotation

    def export_settings(self, file_name: str) -> None:
        """Export camera settings to .npz file.

        :param file_name: Name of the file to which camera settings are saved.
        """
        settings = {
            "orthographic_projection": self.orthographic_projection,
            "fov": self.fov,
            "width": self.width,
            "height": self.height,
            "zoom_sensitivity": self.zoom_sensitivity,
            "distance_from_target": self.distance_from_target,
            "position": self.position.tolist(),
            "up_vector": self.up_vector.tolist(),
            "right_vector": self.right_vector.tolist(),
            "target": self.target.tolist(),
            "translation": self.translation.tolist(),
            "initial_position": self.initial_position.tolist(),
            "initial_up_vector": self.initial_up_vector.tolist(),
            "initial_right_vector": self.initial_right_vector.tolist(),
            "initial_target": self.initial_target.tolist(),
            "last_translation": self.last_translation.tolist(),
            "rotation": self.rotation.tolist(),
            "last_rotation": self.last_rotation.tolist(),
        }
        if not file_name.endswith(".json"):
            file_name += ".json"
        with open(file_name, "w") as file:
            json.dump(settings, file, indent=4)

    def import_settings(self, file_name: str) -> None:
        """Import camera settings from .npz file.

        :param file_name: Name of the file from which camera settings are loaded.
        """
        if not file_name.endswith(".json"):
            # Show warning
            return
        with open(file_name) as file:
            data = json.load(file)
        self.orthographic_projection = data["orthographic_projection"]
        self.fov = data["fov"]
        self.width = data["width"]
        self.height = data["height"]
        self.zoom_sensitivity = data["zoom_sensitivity"]
        self.set_position(
            data["position"],
            data["up_vector"],
            data["right_vector"],
            data["distance_from_target"],
        )
        self.initial_position = pyrr.Vector3(data["initial_position"], dtype=np.float32)
        self.initial_up_vector = pyrr.Vector3(data["initial_up_vector"], dtype=np.float32)
        self.initial_right_vector = pyrr.Vector3(data["initial_right_vector"], dtype=np.float32)
        self.target = pyrr.Vector3(data["target"], dtype=np.float32)
        self.initial_target = pyrr.Vector3(data["initial_target"], dtype=np.float32)
        self.translation = pyrr.Vector3(data["translation"], dtype=np.float32)
        self.last_translation = pyrr.Vector3(data["last_translation"], dtype=np.float32)
        self.rotation = pyrr.Quaternion(data["rotation"])
        self.last_rotation = pyrr.Quaternion(data["last_rotation"])
        self.calculate_projection_matrix()
        self.update()

    def adopt_config(self, other_camera: Camera, custom_geometry: tuple[int, int] | None = None) -> None:
        """Adopt the configuration of another Camera object.

        :param other_camera: the other Camera object
        :param custom_geometry: custom geometry (width, height) for the Camera object
        """
        self.distance_from_target = other_camera.distance_from_target
        self.fov = other_camera.fov
        self.zoom_sensitivity = other_camera.zoom_sensitivity
        if custom_geometry is None:
            self.width = other_camera.width
            self.height = other_camera.height
        else:
            self.width, self.height = custom_geometry[0], custom_geometry[1]
        self.initial_position = other_camera.initial_position
        self.initial_right_vector = other_camera.initial_right_vector
        self.initial_target = other_camera.initial_target
        self.initial_up_vector = other_camera.initial_up_vector
        self.last_rotation = other_camera.last_rotation
        self.last_translation = other_camera.last_translation
        self.position = other_camera.position
        self.right_vector = other_camera.right_vector
        self.rotation = other_camera.rotation
        self.target = other_camera.target
        self.translation = other_camera.translation
        self.up_vector = other_camera.up_vector
        self.orthographic_projection = other_camera.orthographic_projection
        self.calculate_projection_matrix()
        self.update()



import numpy as np
import pyrr


cpdef int select_sphere(float[:] screen_coordinates, float[:] camera_position,
                        float[:,:] view_matrix_inv, float[:,:] projection_matrix_inv, float fov, float aspect_ratio,
                        float[:,:] sphere_origins, float[:] radii):
    """Select the sphere that is closest to the screen coordinates.

    :param screen_coordinates: Screen coordinates of mouse click
    :param camera_position: Position of camera
    :param view_matrix_inv: Inverted view matrix of camera
    :param projection_matrix_inv: Inverted projection matrix of camera
    :param fov: Field of view of camera
    :param aspect_ratio: Aspect ratio of camera
    :param sphere_origins: Origins of spheres
    :param radii: Radii of spheres
    :return: Index of selected sphere or -1 if no sphere was selected
    """

    cdef int i, selected_sphere_index
    cdef float min_distance, distance, ray[3]

    min_distance = -1.0
    selected_sphere_index = -1

    # Calculate ray origin and direction
    ray[0] = screen_coordinates[0] * projection_matrix_inv[1,1] / aspect_ratio
    ray[1] = -screen_coordinates[1] * projection_matrix_inv[1,1]
    ray[2] = -1.0

    # This should not work... But it does, I do not know why :)

    ray = pyrr.matrix44.inverse(view_matrix_inv).dot(np.append(ray, 0.0))[:3]
    ray = ray / np.linalg.norm(ray)

    for i in range(sphere_origins.shape[0]):
        distance = check_sphere_intersect(sphere_origins[i,:], radii[i], camera_position, ray)
        if distance > 0:
            if min_distance < 0 or distance < min_distance:
                min_distance = distance
                selected_sphere_index = i

    return selected_sphere_index



cpdef float check_sphere_intersect(float[:] sphere_origin, float radius,
                                   float[:] ray_origin, float[:] ray_direction):
    """Check if a given ray with a specified origin and direction intersects a sphere.

    :param sphere_origin: Origin of sphere
    :param radius: Radius of sphere
    :param ray_origin: Origin of ray
    :param ray_direction: Direction the ray points in. Need to be normalized!
    :return: Distance to intersection point or -1.0 if no intersection
    """

    cdef float discriminant, dx, dy, dz, a, b, c, q, t1, t2, discriminant_sqrt

    dx = ray_origin[0] - sphere_origin[0]
    dy = ray_origin[1] - sphere_origin[1]
    dz = ray_origin[2] - sphere_origin[2]

    a = 1
    b = 2 * (ray_direction[0]*dx + ray_direction[1]*dy + ray_direction[2]*dz)
    c = dx**2 + dy**2 + dz**2 - radius**2
    discriminant = b**2 - 4 * a * c
    if discriminant < 0:
        return -1.0
    elif discriminant > 0:
        discriminant_sqrt = discriminant**0.5
        q = 1 / (2 * a)
        t1 = (-b + discriminant_sqrt) * q
        t2 = (-b - discriminant_sqrt) * q
        if t1 > t2:
            return t2
        else:
            return t1
    else:
        return -b / (2 * a)



"""Module for the measuring tool dialog."""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from PySide6.QtCore import Qt
from PySide6.QtGui import QBrush, QColor
from PySide6.QtWidgets import QDialog, QHeaderView, QMainWindow, QTableWidget, QTableWidgetItem

from molara.gui.ui_measuring_tool import Ui_measuring_tool

if TYPE_CHECKING:
    from molara.structure.structure import Structure

__copyright__ = "Copyright 2024, Molara"


class MeasurementDialog(QDialog):
    """Dialog for displaying measurements."""

    def __init__(self, parent: QMainWindow) -> None:  # noqa: PLR0915
        """Initialize the measurement dialog.

        :param parent: the MainWindow widget
        """
        super().__init__(
            parent,
        )
        self.main_window = parent

        self.ui = Ui_measuring_tool()
        self.ui.setupUi(self)

        self.ui.info_text.setText("Select / unselect atoms: SHIFT + Left Click on atoms.")
        self.ui.info_text_2.setText("Unselect all atoms: SHIFT + CTRL + Left Click on empty area.")

        # basic setup of the tables
        self.ui.tableDistances.setColumnCount(3)
        self.ui.tableDistances.setRowCount(3)
        self.ui.tableAngles.setRowCount(3)
        self.ui.tableAngles.setColumnCount(1)
        self.ui.tablePositions.setColumnCount(5)
        self.ui.tablePositions.setRowCount(4)

        proto_item = QTableWidgetItem()
        proto_item.setTextAlignment(Qt.AlignRight)
        self.ui.tablePositions.setItemPrototype(proto_item)

        # resize modes: specify how the tables are filled with the columns / rows
        def set_resize_modes(obj: QHeaderView, modes: list) -> None:
            for i, mode in enumerate(modes):
                obj.setSectionResizeMode(i, mode)

        fixed, resize, stretch = QHeaderView.Fixed, QHeaderView.ResizeToContents, QHeaderView.Stretch

        header_positions = self.ui.tablePositions.horizontalHeader()
        set_resize_modes(header_positions, [resize, stretch, resize, resize, resize])
        sidelabels_positions = self.ui.tablePositions.verticalHeader()
        set_resize_modes(sidelabels_positions, [fixed, fixed, fixed, stretch])
        header_distances = self.ui.tableDistances.horizontalHeader()
        set_resize_modes(header_distances, [stretch, stretch, stretch])
        sidelabels_distances = self.ui.tableDistances.verticalHeader()
        set_resize_modes(sidelabels_distances, [stretch, stretch, stretch])
        header_angles = self.ui.tableAngles.horizontalHeader()
        set_resize_modes(header_angles, [stretch])
        sidelabels_angles = self.ui.tableAngles.verticalHeader()
        set_resize_modes(sidelabels_angles, [stretch, stretch, stretch])

        # create labels for the table rows & columns
        def _set_table_labels(
            horizontal: bool,
            obj: QTableWidget,
            labels: list[str],
            colors: list[None] | list[str] | None,
            align_right: bool = True,
        ) -> None:
            _set_item = obj.setHorizontalHeaderItem if horizontal else obj.setVerticalHeaderItem
            if colors is None:
                colors = [None] * len(labels)
            for i, (label_i, color_i) in enumerate(zip(labels, colors)):
                item = QTableWidgetItem(label_i)
                if align_right:
                    item.setTextAlignment(Qt.AlignRight | Qt.AlignVCenter)
                if color_i is not None:
                    brush = QBrush(QColor(color_i))
                    item.setForeground(brush)
                _set_item(i, item)

        def set_horizontal_table_labels(obj: QTableWidget, labels: list[str], colors: list[str] | None = None) -> None:
            horizontal = True
            _set_table_labels(horizontal, obj, labels, colors)

        def set_vertical_table_labels(obj: QTableWidget, labels: list[str], colors: list[str] | None = None) -> None:
            vertical = False
            _set_table_labels(vertical, obj, labels, colors)

        colors = ["#f00", "#0d0", "#00f", "#cc0"]
        atom_labels = [rf"Atom {i+1}" for i in range(len(colors))]

        set_horizontal_table_labels(
            self.ui.tablePositions,
            ["Symbol", "Name", "x / \u00c5", "y / \u00c5", "z / \u00c5"],
        )
        set_vertical_table_labels(self.ui.tablePositions, atom_labels, colors)
        set_horizontal_table_labels(self.ui.tableDistances, atom_labels[1:], colors[1:])
        set_vertical_table_labels(self.ui.tableDistances, atom_labels[:-1], colors[:-1])
        set_horizontal_table_labels(self.ui.tableAngles, ["Angle"])
        set_vertical_table_labels(self.ui.tableAngles, ["\u2222 123", "\u2222 234", "\u2222 1234"])

    def display_metrics(self, structure: Structure, selected_atoms: list) -> None:
        """Display the metrics in the table.

        :param structure: The structure to measure.
        :param selected_atoms: The selected atoms.
        :return:
        """
        self.display_positions_distances(structure, selected_atoms)
        self.display_angles(structure, selected_atoms)
        self.display_dihedral(structure, selected_atoms)

    def display_positions_distances(self, structure: Structure, selected_atoms: list) -> None:
        """Display the distances and positions in the table.

        :param structure: The structure to measure.
        :param selected_atoms: The selected atoms.
        :return:
        """
        for i in range(4):
            if selected_atoms[i] != -1:
                pos = structure.atoms[selected_atoms[i]].position

                items = []
                items.append(QTableWidgetItem(structure.atoms[selected_atoms[i]].symbol))
                items.append(QTableWidgetItem(structure.atoms[selected_atoms[i]].name))
                items.append(QTableWidgetItem(f"{pos[0]:.3f}"))
                items.append(QTableWidgetItem(f"{pos[1]:.3f}"))
                items.append(QTableWidgetItem(f"{pos[2]:.3f}"))

                for item in items:
                    item.setTextAlignment(Qt.AlignRight | Qt.AlignVCenter)
                    self.ui.tablePositions.setItem(i, items.index(item), item)

            else:
                self.ui.tablePositions.setItem(i, 0, QTableWidgetItem(""))
                self.ui.tablePositions.setItem(i, 1, QTableWidgetItem(""))
                self.ui.tablePositions.setItem(i, 2, QTableWidgetItem(""))
                self.ui.tablePositions.setItem(i, 3, QTableWidgetItem(""))
                self.ui.tablePositions.setItem(i, 4, QTableWidgetItem(""))
            for k in range(i + 1, 4):
                if selected_atoms[i] == -1 or selected_atoms[k] == -1:
                    self.ui.tableDistances.setItem(i, k - 1, QTableWidgetItem(""))
                    continue
                d = np.linalg.norm(
                    structure.atoms[selected_atoms[i]].position - structure.atoms[selected_atoms[k]].position,
                )
                item = QTableWidgetItem(f"{d:.3f}" + " \u00c5")
                item.setTextAlignment(Qt.AlignRight | Qt.AlignVCenter)
                self.ui.tableDistances.setItem(i, k - 1, item)

    def display_angles(self, structure: Structure, selected_atoms: list) -> None:
        """Display the angles in the table.

        :param structure: The structure to measure.
        :param selected_atoms: The selected atoms.
        :return:
        """
        for i in range(2):
            if selected_atoms[i] != -1 and selected_atoms[i + 1] != -1 and selected_atoms[i + 2] != -1:
                v1 = structure.atoms[selected_atoms[i]].position - structure.atoms[selected_atoms[i + 1]].position
                v2 = structure.atoms[selected_atoms[i + 2]].position - structure.atoms[selected_atoms[i + 1]].position
                a = np.arccos(
                    np.dot(v1, v2) / (np.linalg.norm(v1) * np.linalg.norm(v2)),
                )
                item = QTableWidgetItem(f"{np.degrees(a):.3f}" + " \u00b0")
                item.setTextAlignment(Qt.AlignRight | Qt.AlignVCenter)
                self.ui.tableAngles.setItem(i, 0, item)
            else:
                self.ui.tableAngles.setItem(i, 0, QTableWidgetItem(""))

    def display_dihedral(self, structure: Structure, selected_atoms: list) -> None:
        """Display the dihedral in the table.

        :param structure: The structure to measure.
        :param selected_atoms: The selected atoms.
        :return:
        """
        atom_id1, atom_id2, atom_id3, atom_id4 = selected_atoms
        if atom_id1 == -1 or atom_id2 == -1 or atom_id3 == -1 or atom_id4 == -1:
            self.ui.tableAngles.setItem(2, 0, QTableWidgetItem(""))
            return

        ab = structure.atoms[atom_id2].position - structure.atoms[atom_id1].position
        bc = structure.atoms[atom_id3].position - structure.atoms[atom_id2].position
        cd = structure.atoms[atom_id4].position - structure.atoms[atom_id3].position
        # explanations on the calculation see https://math.stackexchange.com/a/47084
        nabc = np.cross(ab, bc)
        nbcd = np.cross(bc, cd)
        nabc /= np.linalg.norm(nabc)
        nbcd /= np.linalg.norm(nbcd)
        m1 = np.cross(nabc, bc / np.linalg.norm(bc))
        x = np.dot(nabc, nbcd)
        y = np.dot(m1, nbcd)
        a = np.arctan2(y, x)
        a_deg = np.degrees(a)
        item = QTableWidgetItem(f"{a_deg:.3f}" + " \u00b0")
        item.setTextAlignment(Qt.AlignRight | Qt.AlignVCenter)
        self.ui.tableAngles.setItem(2, 0, item)

    def reject(self) -> None:
        """Close the dialog window."""
        structure_widget = self.main_window.structure_widget
        self.main_window.structure_widget.unselect_all_atoms()
        self.display_metrics(structure_widget.structures[0], structure_widget.measurement_selected_spheres)
        self.setVisible(False)




cpdef double norm(double[:] x):
    cdef double res = 0
    cdef int i
    for i in range(x.shape[0]):
        res += x[i]*x[i]
    res = res**0.5
    return res

cpdef float norm_float(float[:] x):
    cdef float res = 0
    cdef int i
    for i in range(x.shape[0]):
        res += x[i]*x[i]
    res = res**0.5
    return res



"""This module contains the Sphere and Spheres classes."""

import numpy as np
cimport numpy as npc
from molara.tools.mathtools import norm

class Sphere:
    """Creates a Sphere object, containing its vertices and indices.

    :param subdivisions: Number of subdivisions of the sphere.
    """

    def __init__(self, subdivisions: int) -> None:
        """Creates a Sphere object, containing its vertices and indices."""
        self.subdivisions = subdivisions
        vertices, indices = generate_sphere(self.subdivisions)
        self.vertices = vertices
        self.indices = indices


def generate_sphere(
    subdivisions: int,
) -> tuple[np.ndarray, np.ndarray]:
    """Calculates the vertices and indices of a sphere for a given number of subdivisions.

    :param subdivisions: Number of subdivisions of the sphere.
    :returns:
        - **vertices** (numpy.array of numpy.float32) - Vertices in the following order x,y,z,nx,ny,nz,..., where\
         xyz are the cartesian coordinates and nxnynz are the components of the normal\
          vector.
        - **indices** (numpy.array of numpy.uint32) - Gives the connectivity of the vertices.
    """
    vertices = []
    indices = []

    for i in range(subdivisions + 1):
        phi = np.pi * (i / subdivisions - 0.5)
        y = np.sin(phi)

        for j in range(subdivisions * 2 + 1):
            theta = 2 * np.pi * j / (subdivisions * 2)
            x = np.cos(theta) * np.cos(phi)
            z = np.sin(theta) * np.cos(phi)

            normal = np.array([x, y, z])
            normal /= norm(normal)
            vertices.extend(
                [
                    x,
                    y,
                    z,
                    normal[0],
                    normal[1],
                    normal[2],
                ],
            )
            if j < subdivisions * 2 and i < subdivisions:
                p1 = i * (subdivisions * 2 + 1) + j
                p2 = p1 + 1
                p3 = p1 + subdivisions * 2 + 1
                p4 = p3 + 1

                indices.extend([p1, p2, p3, p3, p2, p4])

    return np.array(vertices, dtype=np.float32), np.array(indices, dtype=np.uint32)

def calculate_sphere_model_matrix(npc.ndarray[float, ndim=1] position,
                                  float radius) -> np.ndarray:
    """Calculates the model matrix for a sphere.

    :param position: Position of the sphere.
    :param radius: Radius of the sphere.
    :return: Model matrix of the sphere.
    """

    cdef npc.ndarray[float, ndim=2] translation_matrix
    cdef npc.ndarray[float, ndim=2] scale_matrix
    cdef npc.ndarray[float, ndim=2] model_matrix

    translation_matrix = np.array([[1.,0.,0.,0.],[0.,1.,0.,0.],[0.,0.,1.,0.],[0.,0.,0.,1.]], dtype=np.float32)
    translation_matrix[3, 0:3] = position

    scale_matrix =np.array([[1.,0.,0.,0.],[0.,1.,0.,0.],[0.,0.,1.,0.],[0.,0.,0.,1.]], dtype=np.float32)
    scale_matrix[0,0] = radius
    scale_matrix[1,1] = radius
    scale_matrix[2,2] = radius

    model_matrix = scale_matrix @ translation_matrix
    return np.array(
        [model_matrix],
        dtype=np.float32,
    )



import numpy as np
cimport numpy as npc
from cython.parallel import prange
from cython import boundscheck, exceptval
from cython import nogil

@boundscheck(False)
def calculate_model_matrices(float[:,:,:] translation, float[:,:,:] scale,
                             float[:,:,:] rotation=np.array([[[-1.0]]], dtype=np.float32),
                             bint cylinder=False):

    cdef int n = translation.shape[0], i, j, k, l
    cdef float[:,:,:] model_matrices = np.zeros((n, 4, 4), dtype=np.float32)
    cdef float[:,:] temp = np.zeros((3,3), dtype=np.float32)
    cdef float temp1

    model_matrices[:, 3, 3] = 1.0
    with nogil:
        if rotation[0, 0, 0] == -1:
            for i in prange(n):
                model_matrices[i, 0, 0] = scale[i, 0, 0]
                model_matrices[i, 1, 1] = scale[i, 1, 1]
                model_matrices[i, 2, 2] = scale[i, 2, 2]
                model_matrices[i, 3, 0] = translation[i, 3, 0]
                model_matrices[i, 3, 1] = translation[i, 3, 1]
                model_matrices[i, 3, 2] = translation[i, 3, 2]
        else:
            for i in prange(n):
                if cylinder:
                    if (i % 2) == 0:
                        dot_product_m(rotation[i//2, :, :], scale[i//2, :, :], temp)
                else:
                    dot_product_m(rotation[i, :, :], scale[i, :, :], temp)
                model_matrices[i, :3, :3] = temp[:,:]
                model_matrices[i, 3, 0] = translation[i, 3, 0]
                model_matrices[i, 3, 1] = translation[i, 3, 1]
                model_matrices[i, 3, 2] = translation[i, 3, 2]
    return np.array(
        model_matrices,
        dtype=np.float32,
    )
@exceptval(check=False)
@boundscheck(False)
cdef int dot_product_m(float[:,:] a, float[:,:] b, float[:,:] res) nogil:
    """Calculates the dot product of two matrices.

    :param res: Resulting matrix.
    :param a: First matrix.
    :param b: Second matrix.
    """
    cdef int j, k, l

    for j in prange(3):
        for k in prange(3):
            res[k, j] = 0.0
            for l in range(3):
                res[k, j] = res[k, j] + a[j, l] * b[l, k]

    return 0
@boundscheck(False)
def calculate_translation_matrices(npc.ndarray[float, ndim=2] positions) -> npc.ndarray:
    """Calculates the translation matrix for a sphere.

    :param position: Position of the sphere.
    :return: Translation matrix of the sphere.
    """

    cdef npc.ndarray[float, ndim=3] translation_matrices = np.zeros((positions.shape[0], 4, 4), dtype=np.float32)
    cdef int n = positions.shape[0], i

    with nogil:
        for i in prange(n):
            translation_matrices[i, 0, 0] = 1.0
            translation_matrices[i, 1, 1] = 1.0
            translation_matrices[i, 2, 2] = 1.0
            translation_matrices[i, 3, 3] = 1.0
            translation_matrices[i, 3, 0] = positions[i, 0]
            translation_matrices[i, 3, 1] = positions[i, 1]
            translation_matrices[i, 3, 2] = positions[i, 2]


    return np.array(translation_matrices, dtype=np.float32)

@boundscheck(False)
def calculate_scale_matrices(npc.ndarray[float, ndim=2] scales) -> np.ndarray:
    """Calculates the scale matrix for a sphere.

    :param scales: Scales of the spheres.
    :return: Scale matrix of the sphere.
    """

    cdef npc.ndarray[float, ndim=3] scale_matrices = np.zeros((scales.shape[0], 4, 4), dtype=np.float32)
    cdef int n = scales.shape[0], i

    with nogil:
        for i in prange(n):
            scale_matrices[i, 0, 0] = scales[i, 0]
            scale_matrices[i, 1, 1] = scales[i, 1]
            scale_matrices[i, 2, 2] = scales[i, 2]
            scale_matrices[i, 3, 3] = 1.0

    return np.array(scale_matrices, dtype=np.float32)

@boundscheck(False)
def calculate_rotation_matrices(
    double[:,:] directions,
):
    """Calculates the rotation matrix.

    :param directions: Direction the y-axis of the rotated object should be rotated to.
    :return: Rotation matrices.
    """
    cdef npc.ndarray[float, ndim=3] rotation_matrices = np.zeros((directions.shape[0], 4, 4), dtype=np.float32)
    cdef int n = directions.shape[0], i, j, k
    cdef double[3] rotation_axis
    cdef float rotation_angle, x, y, z, c, s, t, dot_product
    cdef float normalized_direction[3]
    cdef double direction_norm
    cdef float[3] y_axis = [0., 1., 0.]

    with (nogil):
        for i in prange(n):
            direction_norm = (directions[i, 0]**2 + directions[i, 1]**2 + directions[i, 2]**2)**0.5
            normalized_direction[0] = directions[i, 0] / direction_norm
            normalized_direction[1] = directions[i, 1] / direction_norm
            normalized_direction[2] = directions[i, 2] / direction_norm
            dot_product = -(y_axis[0]*normalized_direction[0] + y_axis[1]*normalized_direction[1] + y_axis[2]*normalized_direction[2])
            if dot_product != 1 and dot_product != -1:
                rotation_axis[0] = y_axis[1] * normalized_direction[2] - y_axis[2] * normalized_direction[1]
                rotation_axis[1] = y_axis[2] * normalized_direction[0] - y_axis[0] * normalized_direction[2]
                rotation_axis[2] = y_axis[0] * normalized_direction[1] - y_axis[1] * normalized_direction[0]
                c = dot_product
                s = (rotation_axis[0]**2 + rotation_axis[1]**2 + rotation_axis[2]**2)**0.5
                x = rotation_axis[0] / s
                y = rotation_axis[1] / s
                z = rotation_axis[2] / s
                t = 1 - c

                rotation_matrices[i, 0, 0] = t*x*x + c
                rotation_matrices[i, 0, 1] = t*x*y + s*z
                rotation_matrices[i, 0, 2] = t*x*z - s*y
                rotation_matrices[i, 1, 0] = t*x*y - s*z
                rotation_matrices[i, 1, 1] = t*y*y + c
                rotation_matrices[i, 1, 2] = t*y*z + s*x
                rotation_matrices[i, 2, 0] = t*x*z + s*y
                rotation_matrices[i, 2, 1] = t*y*z - s*x
                rotation_matrices[i, 2, 2] = t*z*z + c
                rotation_matrices[i, 3, 3] = 1.0
            else:
                for j in prange(3):
                    for k in prange(3):
                        if j == k:
                            rotation_matrices[i, j, k] = 1.0
                        else:
                            rotation_matrices[i, j, k] = 0.0

    return np.array(rotation_matrices, dtype=np.float32)
